/* Automatically generated by po2tbl.sed from textutils.pot.  */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "libgettext.h"

const struct _msg_ent _msg_tbl[] = {
  {"", 1},
  {"Try `%s --help' for more information.\n", 2},
  {"Usage: %s [OPTION] [FILE]...\n", 3},
  {"\
Concatenate FILE(s), or standard input, to standard output.\n\
\n\
  -b, --number-nonblank    number nonblank output lines\n\
  -e                       equivalent to -vE\n\
  -n, --number             number all output lines\n\
  -s, --squeeze-blank      never more than one single blank line\n\
  -t                       equivalent to -vT\n\
  -u                       (ignored)\n\
  -v, --show-nonprinting   use ^ and M- notation, save for LFD and TAB\n\
  -A, --show-all           equivalent to -vET\n\
  -E, --show-ends          display $ at end of each line\n\
  -T, --show-tabs          display TAB characters as ^I\n\
      --help               display this help and exit\n\
      --version            output version information and exit\n\
\n\
With no FILE, or when FILE is -, read standard input.\n", 4},
  {"write error", 5},
  {"cannot do ioctl on `%s'", 6},
  {"standard output", 7},
  {"%s: input file is output file", 8},
  {"Usage: %s [OPTION]... [FILE]...\n", 9},
  {"\
Print CRC checksum and byte counts of each FILE.\n\
\n\
  --help      display this help and exit\n\
  --version   output version information and exit\n", 10},
  {"Usage: %s [OPTION]... LEFT_FILE RIGHT_FILE\n", 11},
  {"\
Compare sorted files LEFT_FILE and RIGHT_FILE line by line.\n\
\n\
  -1              suppress lines unique to left file\n\
  -2              suppress lines unique to right file\n\
  -3              suppress lines unique to both files\n\
      --help      display this help and exit\n\
      --version   output version information and exit\n", 12},
  {"virtual memory exhausted", 13},
  {"read error", 14},
  {"input disappeared", 15},
  {"%s: line number out of range", 16},
  {"%s: `%d': line number out of range", 17},
  {" on repetition %d\n", 18},
  {"%s: `%s': match not found", 19},
  {"error in regular expression search", 20},
  {"write error for `%s'", 21},
  {"%s: `+' or `-' expected after delimeter", 22},
  {"%s: integer expected after `%c'", 23},
  {"%s: `}' is required in repeat count", 24},
  {"%s}: integer required between `{' and `}'", 25},
  {"%s: closing delimeter `%c' missing", 26},
  {"%s: invalid regular expression: %s", 27},
  {"%s: invalid pattern", 28},
  {"%s: line number must be greater than zero", 29},
  {"line number `%s' is smaller than preceding line number, %lu", 30},
  {"warning: line number `%s' is the same as preceding line number", 31},
  {"missing conversion specifier in suffix", 32},
  {"invalid conversion specifier in suffix: %c", 33},
  {"invalid conversion specifier in suffix: \\%.3o", 34},
  {"missing %% conversion specification in suffix", 35},
  {"too many %% conversion specifications in suffix", 36},
  {"%s: invalid number", 37},
  {"too few arguments", 38},
  {"Usage: %s [OPTION]... FILE PATTERN...\n", 39},
  {"\
Output pieces of FILE separated by PATTERN(s) to files `xx01', `xx02', ...,\n\
and output byte counts of each piece to standard output.\n\
\n\
  -b, --suffix-format=FORMAT use sprintf FORMAT instead of %%d\n\
  -f, --prefix=PREFIX        use PREFIX instead of `xx'\n\
  -k, --keep-files           do not remove output files on errors\n\
  -n, --digits=DIGITS        use specified number of digits instead of 2\n\
  -s, --quiet, --silent      do not print counts of output file sizes\n\
  -z, --elide-empty-files    remove empty output files\n\
      --help                 display this help and exit\n\
      --version              output version information and exit\n\
\n\
Read standard input if FILE is -.  Each PATTERN may be:\n\
\n\
  INTEGER            copy up to but not including specified line number\n\
  /REGEXP/[OFFSET]   copy up to but not including a matching line\n\
  %%REGEXP%%[OFFSET]   skip to, but not including a matching line\n\
  {INTEGER}          repeat the previous pattern specified number of times\n\
  {*}                repeat the previous pattern as many times as possible\n\
\n\
A line OFFSET is a required `+' or `-' followed by a positive integer.\n", 40},
  {"\
Print selected parts of lines from each FILE to standard output.\n\
\n\
  -b, --bytes=LIST        output only these bytes\n\
  -c, --characters=LIST   output only these characters\n\
  -d, --delimiter=DELIM   use DELIM instead of TAB for field delimiter\n\
  -f, --fields=LIST       output only these fields\n\
  -n                      (ignored)\n\
  -s, --only-delimited    do not print lines not containing delimiters\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n\
\n\
Use one, and only one of -b, -c or -f.  Each LIST is made up of one\n\
range, or many ranges separated by commas.  Each range is one of:\n\
\n\
  N     N'th byte, character or field, counted from 1\n\
  N-    from N'th byte, character or field, to end of line\n\
  N-M   from N'th to M'th (included) byte, character or field\n\
  -M    from first to M'th (included) byte, character or field\n\
\n\
With no FILE, or when FILE is -, read standard input.\n", 41},
  {"invalid byte or field list", 42},
  {"only one type of list may be specified", 43},
  {"missing list of positions", 44},
  {"missing list of fields", 45},
  {"the delimiter must be a single character", 46},
  {"you must specify a list of bytes, characters, or fields", 47},
  {"a delimiter may be specified only when operating on fields", 48},
  {"\
suppressing non-delimited lines makes sense\n\
\tonly when operating on fields", 49},
  {"\
Convert tabs in each FILE to spaces, writing to standard output.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -i, --initial       do not convert TABs after non whitespace\n\
  -t, --tabs=NUMBER   have tabs NUMBER characters apart, not 8\n\
  -t, --tabs=LIST     use comma separated list of explicit tab positions\n\
      --help          display this help and exit\n\
      --version       output version information and exit\n\
\n\
Instead of -t NUMBER or -t LIST, -NUMBER or -LIST may be used.\n", 50},
  {"tab size contains an invalid character", 51},
  {"tab size cannot be 0", 52},
  {"tab sizes must be ascending", 53},
  {"Usage: %s [-DIGITS] [OPTION]... [FILE]...\n", 54},
  {"\
Reformat each paragraph in the FILE(s), writing to standard output.\n\
If no FILE or if FILE is `-', read standard input.\n\
\n\
Mandatory arguments to long options are mandatory for short options too.\n\
  -c, --crown-margin        preserve indentation of first two lines\n\
  -s, --split-only          split long lines, but do not refill\n\
  -t, --tagged-paragraph    indentation of first line different from second\n\
  -u, --uniform-spacing     one space between words, two after sentences\n\
  -w, --width=NUMBER        maximum line width (default of 75 columns)\n\
  -p, --prefix=STRING       combine only lines having STRING as prefix\n\
      --help                display this help and exit\n\
      --version             output version information and exit\n\
\n\
In -wNUMBER, the letter `w' may be omitted.\n", 55},
  {"invalid line number increment: `%s'", 56},
  {"\
Wrap input lines in each FILE (standard input by default), writing to\n\
standard output.\n\
\n\
  -b, --bytes         count bytes rather than columns\n\
  -s, --spaces        break at spaces\n\
  -w, --width=WIDTH   use WIDTH columns instead of 80\n", 57},
  {"invalid number of columns: `%s'", 58},
  {"\
Print first 10 lines of each FILE to standard output.\n\
With more than one FILE, precede each with a header giving the file name.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -c, --bytes=SIZE         print first SIZE bytes\n\
  -n, --lines=NUMBER       print first NUMBER lines instead of first 10\n\
  -q, --quiet, --silent    never print headers giving file names\n\
  -v, --verbose            always print headers giving file names\n\
      --help               display this help and exit\n\
      --version            output version information and exit\n\
\n\
SIZE may have a multiplier suffix: b for 512, k for 1K, m for 1 Meg.\n\
If -VALUE is used as first OPTION, read -c VALUE when one of\n\
multipliers bkm follows concatenated, else read -n VALUE.\n", 59},
  {"standard input", 60},
  {"unrecognized option `-%c'", 61},
  {"invalid number `%s'", 62},
  {"Usage: %s [OPTION]... FILE1 FILE2\n", 63},
  {"\
For each pair of input lines with identical join fields, write a line to\n\
standard output.  The default join field is the first, delimited\n\
by whitespace.  When FILE1 or FILE2 (not both) is -, read standard input.\n\
\n\
  -a SIDE          print unpairable lines coming from file SIDE\n\
  -e EMPTY         replace missing input fields with EMPTY\n\
  -i, --ignore-case ignore differences in case when comparing fields\n\
  -j FIELD         (Obsolescent) equivalent to `-1 FIELD -2 FIELD'\n\
  -j1 FIELD        (Obsolescent) equivalent to `-1 FIELD'\n\
  -j2 FIELD        (Obsolescent) equivalent to `-2 FIELD'\n\
  -1 FIELD         join on this FIELD of file 1\n\
  -2 FIELD         join on this FIELD of file 2\n\
  -o FORMAT        obey FORMAT while constructing output line\n\
  -t CHAR          use CHAR as input and output field separator\n\
  -v SIDE          like -a SIDE, but suppress joined output lines\n\
  --help           display this help and exit\n\
  --version        output version information and exit\n\
\n\
Unless -t CHAR is given, leading blanks separate fields and are ignored,\n\
else fields are separated by CHAR.  Any FIELD is a field number counted\n\
from 1.  FORMAT is one or more comma or blank separated specifications,\n\
each being `SIDE.FIELD' or `0'.  Default FORMAT outputs the join field,\n\
the remaining fields from FILE1, the remaining fields from FILE2, all\n\
separated by CHAR.\n", 64},
  {"invalid field specifier: `%s'", 65},
  {"invalid field number: `%s'", 66},
  {"invalid file number in field spec: `%s'", 67},
  {"invalid field number for file 1: `%s'", 68},
  {"invalid field number for file 2: `%s'", 69},
  {"too many non-option arguments", 70},
  {"too few non-option arguments", 71},
  {"both files cannot be standard input", 72},
  {"\
Usage: %s [OPTION] [FILE]...\n\
  or:  %s [OPTION] --check [FILE]\n\
  or:  %s [OPTION] --string=STRING ...\n\
Print or check MD5 checksums.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -b, --binary            read files in binary mode\n\
  -t, --text              read files in text mode (default)\n\
  -c, --check             check MD5 sums against given list\n\
\n\
The following two options are useful only when verifying checksums:\n\
      --status            don't output anything, status code shows success\n\
  -w, --warn              warn about improperly formated MD5 checksum lines\n\
\n\
      --string=STRING     compute checksum for STRING\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n\
\n\
The sums are computed as described in RFC 1321.  When checking, the input\n\
should be a former output of this program.  The default mode is to print\n\
a line with checksum, a character indicating type (`*' for binary, ` ' for\n\
text), and name for each FILE.\n", 73},
  {"%s: %lu: improperly formatted MD5 checksum line", 74},
  {"%s: FAILED open or read\n", 75},
  {"FAILED", 76},
  {"OK", 77},
  {"%s: read error", 78},
  {"%s: no properly formatted MD5 checksum lines found", 79},
  {"WARNING: %d of %d listed %s could not be read\n", 80},
  {"file", 81},
  {"files", 82},
  {"WARNING: %d of %d computed %s did NOT match", 83},
  {"checksum", 84},
  {"checksums", 85},
  {"the --string and --check options are mutually exclusive", 86},
  {"the --status option is meaningful only when verifying checksums", 87},
  {"the --warn option is meaningful only when verifying checksums", 88},
  {"no files may be specified when using --string", 89},
  {"only one argument may be specified when using --check", 90},
  {"\
Write each FILE to standard output, with line numbers added.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -b, --body-numbering=STYLE      use STYLE for numbering body lines\n\
  -d, --section-delimiter=CC      use CC for separating logical pages\n\
  -f, --footer-numbering=STYLE    use STYLE for numbering footer lines\n\
  -h, --header-numbering=STYLE    use STYLE for numbering header lines\n\
  -i, --page-increment=NUMBER     line number increment at each line\n\
  -l, --join-blank-lines=NUMBER   group of NUMBER empty lines counted as \
one\n\
  -n, --number-format=FORMAT      insert line numbers according to FORMAT\n\
  -p, --no-renumber               do not reset line numbers at logical \
pages\n\
  -s, --number-separator=STRING   add STRING after (possible) line number\n\
  -v, --first-page=NUMBER         first line number on each logical page\n\
  -w, --number-width=NUMBER       use NUMBER columns for line numbers\n\
      --help                      display this help and exit\n\
      --version                   output version information and exit\n\
\n\
By default, selects -v1 -i1 -l1 -sTAB -w6 -nrn -hn -bt -fn.  CC are\n\
two delimiter characters for separating logical pages, a missing\n\
second character implies :.  Type \\\\ for \\.  STYLE is one of:\n\
\n\
  a         number all lines\n\
  t         number only nonempty lines\n\
  n         number no lines\n\
  pREGEXP   number only lines that contain a match for REGEXP\n\
\n\
FORMAT is one of:\n\
\n\
  ln   left justified, no leading zeros\n\
  rn   right justified, no leading zeros\n\
  rz   right justified, leading zeros\n\
\n", 91},
  {"invalid starting line number: `%s'", 92},
  {"invalid number of blank lines: `%s'", 93},
  {"invalid line number field width: `%s'", 94},
  {"\
Usage: %s [OPTION]... [FILE]...\n\
  or:  %s --traditional [FILE] [[+]OFFSET [[+]LABEL]]\n", 95},
  {"\
Write an unambiguous representation, octal bytes by default, of FILE\n\
to standard output.  With no FILE, or when FILE is -, read standard input.\n\
\n\
  -A, --address-radix=RADIX   decide how file offsets are printed\n\
  -N, --read-bytes=BYTES      limit dump to BYTES input bytes per file\n\
  -j, --skip-bytes=BYTES      skip BYTES input bytes first on each file\n\
  -s, --strings[=BYTES]       output strings of at least BYTES graphic \
chars\n\
  -t, --format=TYPE           select output format or formats\n\
  -v, --output-duplicates     do not use * to mark line suppression\n\
  -w, --width[=BYTES]         output BYTES bytes per output line\n\
      --traditional           accept arguments in pre-POSIX form\n\
      --help                  display this help and exit\n\
      --version               output version information and exit\n\
\n\
Pre-POSIX format specifications may be intermixed, they accumulate:\n\
  -a   same as -t a,  select named characters\n\
  -b   same as -t oC, select octal bytes\n\
  -c   same as -t c,  select ASCII characters or backslash escapes\n\
  -d   same as -t u2, select unsigned decimal shorts\n\
  -f   same as -t fF, select floats\n\
  -h   same as -t x2, select hexadecimal shorts\n\
  -i   same as -t d2, select decimal shorts\n\
  -l   same as -t d4, select decimal longs\n\
  -o   same as -t o2, select octal shorts\n\
  -x   same as -t x2, select hexadecimal shorts\n", 96},
  {"\
\n\
For older syntax (second call format), OFFSET means -j OFFSET.  LABEL\n\
is the pseudo-address at first byte printed, incremented when dump is\n\
progressing.  For OFFSET and LABEL, a 0x or 0X prefix indicates\n\
hexadecimal, suffixes maybe . for octal and b multiply by 512.\n\
\n\
TYPE is made up of one or more of these specifications:\n\
\n\
  a          named character\n\
  c          ASCII character or backslash escape\n\
  d[SIZE]    signed decimal, SIZE bytes per integer\n\
  f[SIZE]    floating point, SIZE bytes per integer\n\
  o[SIZE]    octal, SIZE bytes per integer\n\
  u[SIZE]    unsigned decimal, SIZE bytes per integer\n\
  x[SIZE]    hexadecimal, SIZE bytes per integer\n\
\n\
SIZE is a number.  For TYPE in doux, SIZE may also be C for\n\
sizeof(char), S for sizeof(short), I for sizeof(int) or L for\n\
sizeof(long).  If TYPE is f, SIZE may also be F for sizeof(float), D\n\
for sizeof(double) or L for sizeof(long double).\n\
\n\
RADIX is d for decimal, o for octal, x for hexadecimal or n for none.\n\
BYTES is hexadecimal with 0x or 0X prefix, it is multiplied by 512\n\
with b suffix, by 1024 with k and by 1048576 with m.  -s without a\n\
number implies 3.  -w without a number implies 32.  By default, od\n\
uses -A o -t d2 -w 16.\n", 97},
  {"invalid type string `%s'", 98},
  {"\
invalid type string `%s';\n\
this system doesn't provide a %lu-byte integral type", 99},
  {"\
invalid type string `%s';\n\
this system doesn't provide a %lu-byte floating point type", 100},
  {"invalid character `%c' in type string `%s'", 101},
  {"cannot skip past end of combined input", 102},
  {"old-style offset", 103},
  {"invalid output address radix `%c'; it must be one character from [doxn]", 104},
  {"skip argument", 105},
  {"limit argument", 106},
  {"\
specified number of bytes `%s' is larger than the maximum\n\
representable value of type off_t", 107},
  {"minimum string length", 108},
  {"width specification", 109},
  {"no type may be specified when dumping strings", 110},
  {"invalid second operand in compatibility mode `%s'", 111},
  {"in compatibility mode the last 2 arguments must be offsets", 112},
  {"in compatibility mode there may be no more than 3 arguments", 113},
  {"warning: invalid width %lu; using %d instead", 114},
  {"%d: fmt=\"%s\" width=%d\n", 115},
  {"standard input is closed", 116},
  {"\
Write lines consisting of the sequentially corresponding lines from\n\
each FILE, separated by TABs, to standard output.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -d, --delimiters=LIST   reuse characters from LIST instead of TABs\n\
  -s, --serial            paste one file at a time instead of in parallel\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n\
\n", 117},
  {"`+' requires a numeric argument", 118},
  {"%s: extra characters in the argument to the `-s' option: `%s'\n", 119},
  {"Cannot specify number of columns when printing in parallel.", 120},
  {"Cannot specify both printing across and printing in parallel.", 121},
  {"%s: extra characters in the argument to the `-%c' option: `%s'\n", 122},
  {"page width too narrow", 123},
  {"%s %s  %s Page", 124},
  {"\
Paginate or columnate FILE(s) for printing.\n\
\n\
  +PAGE             begin printing with page PAGE\n\
  -COLUMN           produce COLUMN-column output and print columns down\n\
  -F, -f            simulate formfeed with newlines on output\n\
  -a                print columns across rather than down\n\
  -b                balance columns on the last page\n\
  -c                use hat notation (^G) and octal backslash notation\n\
  -d                double space the output\n\
  -e[CHAR[WIDTH]]   expand input CHARs (TABs) to tab WIDTH (8)\n\
  -h HEADER         use HEADER instead of filename in page headers\n\
  -i[CHAR[WIDTH]]   replace spaces with CHARs (TABs) to tab WIDTH (8)\n\
  -l PAGE_LENGTH    set the page length to PAGE_LENGTH (66) lines\n\
  -m                print all files in parallel, one in each column\n\
  -n[SEP[DIGITS]]   number lines, use DIGITS (5) digits, then SEP (TAB)\n\
  -o MARGIN         offset each line with MARGIN spaces (do not affect -w)\n\
  -r                inhibit warning when a file cannot be opened\n\
  -s[SEP]           separate columns by character SEP (TAB)\n\
  -t                inhibit 5-line page headers and trailers\n\
  -v                use octal backslash notation\n\
  -w PAGE_WIDTH     set page width to PAGE_WIDTH (72) columns\n\
      --help        display this help and exit\n\
      --version     output version information and exit\n\
\n\
-t implied by -l N when N < 10.  Without -s, columns are separated by\n\
spaces.  With no FILE, or when FILE is -, read standard input.\n", 125},
  {"\
Write sorted concatenation of all FILE(s) to standard output.\n\
\n\
  +POS1 [-POS2]    start a key at POS1, end it before POS2\n\
  -M               compare (unknown) < `JAN' < ... < `DEC', imply -b\n\
  -T DIRECT        use DIRECT for temporary files, not $TMPDIR or %s\n\
  -b               ignore leading blanks in sort fields or keys\n\
  -c               check if given files already sorted, do not sort\n\
  -d               consider only [a-zA-Z0-9 ] characters in keys\n\
  -f               fold lower case to upper case characters in keys\n\
  -g               compare according to general numerical value, imply -b\n\
  -i               consider only [\\040-\\0176] characters in keys\n\
  -k POS1[,POS2]   same as +POS1 [-POS2], but all positions counted from 1\n\
  -m               merge already sorted files, do not sort\n\
  -n               compare according to string numerical value, imply -b\n\
  -o FILE          write result on FILE instead of standard output\n\
  -r               reverse the result of comparisons\n\
  -s               stabilize sort by disabling last resort comparison\n\
  -t SEP           use SEParator instead of non- to whitespace transition\n\
  -u               with -c, check for strict ordering\n\
  -u               with -m, only output the first of an equal sequence\n\
  -z               end lines with 0 byte, not newline, for find -print0\n\
      --help       display this help and exit\n\
      --version    output version information and exit\n\
\n\
POS is F[.C][OPTS], where F is the field number and C the character\n\
position in the field, both counted from zero.  OPTS is made up of one\n\
or more of Mbdfinr, this effectively disable global -Mbdfinr settings\n\
for that key.  If no key given, use the entire line as key.  With no\n\
FILE, or when FILE is -, read standard input.\n", 126},
  {"flushing file", 127},
  {"error closing file", 128},
  {"%s: disorder on %s\n", 129},
  {"invalid field specification `%s'", 130},
  {"\
when using the old-style +POS and -POS key specifiers,\n\
the +POS specifier must come first", 131},
  {"option `-k' requires an argument", 132},
  {"the starting field number argument to the `-k' option must be positive", 133},
  {"starting field spec has `.' but lacks following character offset", 134},
  {"\
starting field character offset argument to the `-k' option\n\
must be positive", 135},
  {"field specification has `,' but lacks following field spec", 136},
  {"ending field number argument to the `-k' option must be positive", 137},
  {"ending field spec has `.' but lacks following character offset", 138},
  {"option `-o' requires an argument", 139},
  {"option `-t' requires an argument", 140},
  {"option `-T' requires an argument", 141},
  {"%s: unrecognized option `-%c'\n", 142},
  {"%s: write error", 143},
  {"Usage: %s [OPTION] [INPUT [PREFIX]]\n", 144},
  {"\
Output fixed-size pieces of INPUT to PREFIXaa, PREFIXab, ...; default\n\
PREFIX is `x'.  With no INPUT, or when INPUT is -, read standard input.\n\
\n\
  -C, --line-bytes=SIZE   put at most SIZE bytes of lines per output file\n\
  -b, --bytes=SIZE        put SIZE bytes per output file\n\
  -l, --lines=NUMBER      put NUMBER lines per output file\n\
      --verbose           print a diagnostic to standard error just\n\
\t\t\t    before each output file is opened\n\
  -NUMBER                 same as -l NUMBER\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n\
\n\
SIZE may have a multiplier suffix: b for 512, k for 1K, m for 1 Meg.\n", 145},
  {"creating file `%s'\n", 146},
  {"cannot split in more than one way", 147},
  {"invalid number of bytes", 148},
  {"invalid number of lines", 149},
  {"invalid number", 150},
  {"too many arguments", 151},
  {"\
Print checksum and block counts for each FILE.\n\
\n\
  -r              defeat -s, use BSD sum algorithm, use 1K blocks\n\
  -s, --sysv      use System V sum algorithm, use 512 bytes blocks\n\
      --help      display this help and exit\n\
      --version   output version information and exit\n\
\n\
With no FILE, or when FILE is -, read standard input.\n", 152},
  {"\
Write each FILE to standard output, last line first.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -b, --before             attach the separator before instead of after\n\
  -r, --regex              interpret the separator as a regular expression\n\
  -s, --separator=STRING   use STRING as the separator instead of newline\n\
      --help               display this help and exit\n\
      --version            output version information and exit\n", 153},
  {"separator cannot be empty", 154},
  {"\
Print last 10 lines of each FILE to standard output.\n\
With more than one FILE, precede each with a header giving the file name.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -c, --bytes=N            output the last N bytes\n\
  -f, --follow             output appended data as the file grows\n\
  -n, --lines=N            output the last N lines, instead of last 10\n\
  -q, --quiet, --silent    never output headers giving file names\n\
  -v, --verbose            always output headers giving file names\n\
      --help               display this help and exit\n\
      --version            output version information and exit\n\
\n\
If the first character of N (the number of bytes or lines) is a `+',\n\
print beginning with the Nth item from the start of each file, otherwise,\n\
print the last N items in the file.  N may have a multiplier suffix:\n\
b for 512, k for 1024, m for 1048576 (1 Meg).  A first OPTION of -VALUE\n\
or +VALUE is treated like -n VALUE or -n +VALUE unless VALUE has one of\n\
the [bkm] suffix multipliers, in which case it is treated like -c VALUE\n\
or -c +VALUE.\n", 155},
  {"file truncated", 156},
  {"standard input: cannot follow end of non-regular file", 157},
  {"%s: cannot follow end of non-regular file", 158},
  {"argument", 159},
  {"number of lines", 160},
  {"number of bytes", 161},
  {"Usage: %s [OPTION]... SET1 [SET2]\n", 162},
  {"\
Translate, squeeze, and/or delete characters from standard input,\n\
writing to standard output.\n\
\n\
  -c, --complement        first complement SET1\n\
  -d, --delete            delete characters in SET1, do not translate\n\
  -s, --squeeze-repeats   replace sequence of characters with one\n\
  -t, --truncate-set1     first truncate SET1 to length of SET2\n\
      --help              display this help and exit\n\
      --version           output version information and exit\n", 163},
  {"\
\n\
SETs are specified as strings of characters.  Most represent themselves.\n\
Interpreted sequences are:\n\
\n\
  \\NNN            character with octal value NNN (1 to 3 octal digits)\n\
  \\\\              backslash\n\
  \\a              audible BEL\n\
  \\b              backspace\n\
  \\f              form feed\n\
  \\n              new line\n\
  \\r              return\n\
  \\t              horizontal tab\n\
  \\v              vertical tab\n\
  CHAR1-CHAR2     all characters from CHAR1 to CHAR2 in ascending order\n\
  [CHAR1-CHAR2]   same as CHAR1-CHAR2, if both SET1 and SET2 use this\n\
  [CHAR*]         in SET2, copies of CHAR until length of SET1\n\
  [CHAR*REPEAT]   REPEAT copies of CHAR, REPEAT octal if starting with 0\n\
  [:alnum:]       all letters and digits\n\
  [:alpha:]       all letters\n\
  [:blank:]       all horizontal whitespace\n\
  [:cntrl:]       all control characters\n\
  [:digit:]       all digits\n\
  [:graph:]       all printable characters, not including space\n\
  [:lower:]       all lower case letters\n\
  [:print:]       all printable characters, including space\n\
  [:punct:]       all punctuation characters\n\
  [:space:]       all horizontal or vertical whitespace\n\
  [:upper:]       all upper case letters\n\
  [:xdigit:]      all hexadecimal digits\n\
  [=CHAR=]        all characters which are equivalent to CHAR\n", 164},
  {"\
\n\
Translation occurs if -d is not given and both SET1 and SET2 appear.\n\
-t may be used only when translating.  SET2 is extended to length of\n\
SET1 by repeating its last character as necessary.  Excess characters\n\
of SET2 are ignored.  Only [:lower:] and [:upper:] are guaranteed to\n\
expand in ascending order; used in SET2 while translating, they may\n\
only be used in pairs to specify case conversion.  -s uses SET1 if not\n\
translating nor deleting; else squeezing uses SET2 and occurs after\n\
translation or deletion.\n", 165},
  {"\
warning: the ambiguous octal escape \\%c%c%c is being\n\
\tinterpreted as the 2-byte sequence \\0%c%c, `%c'", 166},
  {"invalid backslash escape at end of string", 167},
  {"invalid backslash escape `\\%c'", 168},
  {"range-endpoints of `%s-%s' are in reverse collating sequence order", 169},
  {"invalid repeat count `%s' in [c*n] construct", 170},
  {"invalid character class `%s'", 171},
  {"%s: equivalence class operand must be a single character", 172},
  {"the [c*] repeat construct may not appear in string1", 173},
  {"only one [c*] repeat construct may appear in string2", 174},
  {"[=c=] expressions may not appear in string2 when translating", 175},
  {"when not truncating set1, string2 must be non-empty", 176},
  {"\
when translating with complemented character classes,\n\
string2 must map all characters in the domain to one", 177},
  {"\
when translating, the only character classes that may appear in\n\
\tstring2 are `upper' and `lower'", 178},
  {"the [c*] construct may appear in string2 only when translating", 179},
  {"two strings must be given when translating", 180},
  {"two strings must be given when both deleting and squeezing repeats", 181},
  {"only one string may be given when deleting without squeezing repeats", 182},
  {"at least one string must be given when squeezing repeats", 183},
  {"misaligned or mismatched upper and/or lower classes", 184},
  {"\
Convert spaces in each FILE to tabs, writing to standard output.\n\
With no FILE, or when FILE is -, read standard input.\n\
\n\
  -a, --all           convert all whitespace, instead of initial whitespace\n\
  -t, --tabs=NUMBER   have tabs NUMBER characters apart instead of 8\n\
  -t, --tabs=LIST     use comma separated list of explicit tab positions\n\
      --help          display this help and exit\n\
      --version       output version information and exit\n\
\n\
Instead of -t NUMBER or -t LIST, -NUMBER or -LIST may be used.\n", 185},
  {"Usage: %s [OPTION]... [INPUT [OUTPUT]]\n", 186},
  {"\
Discard all but one of successive identical lines from INPUT (or\n\
standard input), writing to OUTPUT (or standard output).\n\
\n\
  -c, --count           prefix lines by the number of occurrences\n\
  -d, --repeated        only print duplicate lines\n\
  -f, --skip-fields=N   avoid comparing the first N fields\n\
  -i, --ignore-case     ignore differences in case when comparing\n\
  -s, --skip-chars=N    avoid comparing the first N characters\n\
  -u, --unique          only print unique lines\n\
  -w, --check-chars=N   compare no more than N characters in lines\n\
  -N                    same as -f N\n\
  +N                    same as -s N\n\
      --help            display this help and exit\n\
      --version         output version information and exit\n\
\n\
A field is a run of whitespace, than non-whitespace characters.\n\
Fields are skipped before chars.\n", 187},
  {"error reading %s", 188},
  {"error writing %s", 189},
  {"invalid number of fields to skip: `%s'", 190},
  {"invalid number of bytes to skip: `%s'", 191},
  {"invalid number of bytes to compare: `%s'", 192},
  {"\
Print line, word, and byte counts for each FILE, and a total line if\n\
more than one FILE is specified.  With no FILE, or when FILE is -,\n\
read standard input.\n\
  -l, --lines            print the newline counts\n\
  -w, --words            print the word counts\n\
  -c, --bytes, --chars   print the byte counts\n\
      --help             display this help and exit\n\
      --version          output version information and exit\n", 193},
  {"total", 194},
};

int _msg_tbl_length = 194;
