/* Definitions for GNU textutils, processed by autoheader.
   Copyright (C) 1995 Free Software Foundation, Inc.
*/

/* Define to 1 if NLS is requested.  */
#undef ENABLE_NLS

/* Define as 1 if you have catgets and don't want to use GNU gettext.  */
#undef HAVE_CATGETS

/* Define as 1 if you have gettext and don't want to use GNU gettext.  */
#undef HAVE_GETTEXT

/* Define if your locale.h file contains LC_MESSAGES.  */
#undef HAVE_LC_MESSAGES

/* Define to 1 if you have the stpcpy function.  */
#undef HAVE_STPCPY

/* Define to the name of the distribution.  */
#undef PACKAGE

/* The concatenation of the strings PACKAGE, "-", and VERSION.  */
#undef PACKAGE_VERSION

/* Define to 1 if ANSI function prototypes are usable.  */
#undef PROTOTYPES

/* Define to the version of the distribution.  */
#undef VERSION

/* Define to 1 if GNU regex should be used instead of GNU rx.  */
#undef WITH_REGEX

